@@ -2,6 +2,7 @@ import collections
 import copy
 import operator
 import os
+import re
 
 import sphinx.environment
 import sphinx.util
@@ -51,7 +52,7 @@ def _expand_wildcard_placeholder(original_module, originals_map, placeholder):
                 msg = "Invalid __all__ entry {0} in {1}".format(
                     name, original_module["name"]
                 )
-                LOGGER.warning(msg)
+                LOGGER.warning(msg, type="autoapi", subtype="python_import_resolution")
                 continue
 
             originals.append(originals_map[name])
@@ -106,7 +107,7 @@ def _resolve_module_placeholders(modules, module_name, visit_path, resolved):
             msg = "Cannot resolve cyclic import: {0}, {1}".format(
                 ", ".join(visit_path), imported_from
             )
-            LOGGER.warning(msg)
+            LOGGER.warning(msg, type="autoapi", subtype="python_import_resolution")
             module["children"].remove(child)
             children.pop(child["name"])
             continue
@@ -115,7 +116,7 @@ def _resolve_module_placeholders(modules, module_name, visit_path, resolved):
             msg = "Cannot resolve import of unknown module {0} in {1}".format(
                 imported_from, module_name
             )
-            LOGGER.warning(msg)
+            LOGGER.warning(msg, type="autoapi", subtype="python_import_resolution")
             module["children"].remove(child)
             children.pop(child["name"])
             continue
@@ -143,7 +144,7 @@ def _resolve_module_placeholders(modules, module_name, visit_path, resolved):
             msg = "Cannot resolve import of {0} in {1}".format(
                 child["original_path"], module_name
             )
-            LOGGER.warning(msg)
+            LOGGER.warning(msg, type="autoapi", subtype="python_import_resolution")
             module["children"].remove(child)
             children.pop(child["name"])
             continue
@@ -196,6 +197,28 @@ def _resolve_placeholder(placeholder, original):
     placeholder.update(new)
 
 
+def _link_objs(value):
+    result = ""
+
+    delims = r"(\s*[\[\]\(\),]\s*)"
+    delims_re = re.compile(delims)
+    sub_targets = re.split(delims, value.strip())
+
+    for sub_target in sub_targets:
+        sub_target = sub_target.strip()
+        if delims_re.match(sub_target):
+            result += f"{sub_target}"
+            if sub_target.endswith(","):
+                result += " "
+            else:
+                result += "\\ "
+        elif sub_target:
+            result += f":py:obj:`{sub_target}`\\ "
+
+    # Strip off the extra "\ "
+    return result[:-2]
+
+
 class PythonSphinxMapper(SphinxMapperBase):
 
     """Auto API domain handler for Python
@@ -223,6 +246,7 @@ class PythonSphinxMapper(SphinxMapperBase):
     def __init__(self, app, template_dir=None, url_root=None):
         super(PythonSphinxMapper, self).__init__(app, template_dir, url_root)
 
+        self.jinja_env.filters["link_objs"] = _link_objs
         self._use_implicit_namespace = (
             self.app.config.autoapi_python_use_implicit_namespaces
         )
@@ -294,8 +318,12 @@ class PythonSphinxMapper(SphinxMapperBase):
                 parsed_data = Parser().parse_file(path)
             return parsed_data
         except (IOError, TypeError, ImportError):
-            LOGGER.warning("Unable to read file: {0}".format(path))
             LOGGER.debug("Reason:", exc_info=True)
+            LOGGER.warning(
+                "Unable to read file: {0}".format(path),
+                type="autoapi",
+                subtype="not_readable",
+            )
         return None
 
     def _resolve_placeholders(self):
@@ -335,6 +363,7 @@ class PythonSphinxMapper(SphinxMapperBase):
         try:
             cls = self._OBJ_MAP[data["type"]]
         except KeyError:
+            # this warning intentionally has no (sub-)type
             LOGGER.warning("Unknown type: %s" % data["type"])
         else:
             obj = cls(
@@ -343,7 +372,7 @@ class PythonSphinxMapper(SphinxMapperBase):
                 options=self.app.config.autoapi_options,
                 jinja_env=self.jinja_env,
                 app=self.app,
-                **kwargs
+                **kwargs,
             )
             obj.url_root = self.url_root
 
@@ -370,6 +399,17 @@ class PythonSphinxMapper(SphinxMapperBase):
             yield obj
 
     def _record_typehints(self, obj):
-        if isinstance(obj, (PythonClass, PythonFunction, PythonMethod)):
+        if isinstance(
+            obj, (PythonClass, PythonFunction, PythonMethod)
+        ) and not obj.obj.get("overloads"):
+            obj_annotations = {}
+            for _, name, annotation, _ in obj.obj["args"]:
+                if name and annotation:
+                    obj_annotations[name] = annotation
+
+            return_annotation = obj.obj.get("return_annotation")
+            if return_annotation:
+                obj_annotations["return"] = return_annotation
+
             annotations = self.app.env.temp_data.setdefault("annotations", {})
-            annotations[obj.id] = obj.obj["annotations"]
+            annotations[obj.id] = obj_annotations
